"use strict";
var __create = Object.create;
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __getProtoOf = Object.getPrototypeOf;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toESM = (mod, isNodeMode, target) => (target = mod != null ? __create(__getProtoOf(mod)) : {}, __copyProps(
  isNodeMode || !mod || !mod.__esModule ? __defProp(target, "default", { value: mod, enumerable: true }) : target,
  mod
));
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);

// src/index.ts
var src_exports = {};
__export(src_exports, {
  BaseNeuralNetwork: () => BaseNeuralNetwork,
  DeepLearningClassification: () => DeepLearningClassification,
  DeepLearningRegression: () => DeepLearningRegression,
  FeatureEmbedding: () => FeatureEmbedding,
  LSTMMultivariateTimeSeries: () => LSTMMultivariateTimeSeries,
  LSTMTimeSeries: () => LSTMTimeSeries,
  LogisticRegression: () => LogisticRegression,
  MachineLearningLinearRegression: () => MachineLearningLinearRegression,
  MachineLearningLogisticRegression: () => MachineLearningLogisticRegression,
  MachineLearningModelInterface: () => MachineLearningModelInterface,
  MultipleLinearRegression: () => MultipleLinearRegression,
  TensorScriptModelInterface: () => TensorScriptModelInterface,
  TextEmbedding: () => TextEmbedding,
  createModelFitCallback: () => createModelFitCallback,
  getBackend: () => getBackend,
  getScikit: () => getScikit,
  setBackend: () => setBackend,
  setScikit: () => setScikit
});
module.exports = __toCommonJS(src_exports);

// src/model_interface.ts
var import_tfjs_layers = require("@tensorflow/tfjs-layers");

// src/tensorflow_singleton.ts
var tf = null;
function setBackend(tfInput) {
  tf = tfInput;
}
function getBackend() {
  if (tf === null) {
    throw Error(`
============================
Howdy \u{1F44B}\u{1F44B}. Looks like you are running @jsonstack/model but you haven't set a Tensorflow backend. 
To do so, simply import (or require) your tensorflow library, and call setBackend like so,

import * as tf from '@tensorflow/tfjs';
import * as jsm from '@jsonstack/model';
jsm.setBackend(tf);

That will let @jsonstack/model know you wish to use a tensorflow library to perform your calculations.
============================
    `);
  }
  return tf;
}
function createModelFitCallback(callbackFunctions) {
  const tf2 = getBackend();
  if (Array.isArray(callbackFunctions)) {
    return callbackFunctions;
  } else if (callbackFunctions) {
    return Object.keys(callbackFunctions).map((callbackFunctionName) => {
      return new tf2.CustomCallback({ [callbackFunctionName]: callbackFunctions[callbackFunctionName] });
    });
  } else
    return [];
}

// src/model_interface.ts
var LambdaLayer = class extends import_tfjs_layers.layers.Layer {
  name;
  lambdaFunction;
  lambdaOutputShape;
  constructor(config) {
    super(config);
    if (config.name === void 0) {
      config.name = (+new Date() * Math.random()).toString(36);
    }
    this.name = config.name;
    this.lambdaFunction = config.lambdaFunction;
    if (config.lambdaOutputShape)
      this.lambdaOutputShape = config.lambdaOutputShape;
  }
  call(input, kwargs) {
    let tf2 = getBackend();
    return tf2.tidy(() => {
      let result = new Array();
      result = new Function("input", "tf", this.lambdaFunction)(input, tf2);
      return result;
    });
  }
  computeOutputShape(inputShape) {
    if (this.lambdaOutputShape === void 0) {
      return inputShape[0];
    } else {
      return this.lambdaOutputShape;
    }
  }
  getConfig() {
    const config = {
      ...super.getConfig(),
      lambdaFunction: this.lambdaFunction,
      lambdaOutputShape: this.lambdaOutputShape
    };
    return config;
  }
  static get className() {
    return "LambdaLayer";
  }
};
var TensorScriptModelInterface = class {
  type;
  settings;
  model;
  tokenizer;
  tf;
  trained;
  compiled;
  reshape;
  getInputShape;
  xShape;
  yShape;
  layers;
  getTimeseriesShape;
  loss;
  constructor(options = {}, properties = {}) {
    let tf2 = getBackend();
    this.type = "ModelInterface";
    this.settings = Object.assign({}, options);
    this.model = properties.model;
    this.tf = properties.tf || tf2;
    this.trained = false;
    this.compiled = false;
    this.reshape = TensorScriptModelInterface.reshape;
    this.getInputShape = TensorScriptModelInterface.getInputShape;
    if (this.tf && this.tf.serialization && this.tf.serialization.registerClass)
      this.tf.serialization.registerClass(LambdaLayer);
    return this;
  }
  static reshape(array, shape) {
    const flatArray = flatten(array);
    function product(arr) {
      return arr.reduce((prev, curr) => prev * curr);
    }
    if (!Array.isArray(array) || !Array.isArray(shape)) {
      throw new TypeError("Array expected");
    }
    if (shape.length === 0) {
      throw new DimensionError(0, product(size(array)), "!=");
    }
    let newArray;
    let totalSize = 1;
    const rows = shape[0];
    for (let sizeIndex = 0; sizeIndex < shape.length; sizeIndex++) {
      totalSize *= shape[sizeIndex];
    }
    if (flatArray.length !== totalSize) {
      throw new DimensionError(
        product(shape),
        product(size(array)),
        "!="
      );
    }
    try {
      newArray = _reshape(flatArray, shape);
    } catch (e) {
      if (e instanceof DimensionError) {
        throw new DimensionError(
          product(shape),
          product(size(array)),
          "!="
        );
      }
      throw e;
    }
    if (newArray.length !== rows)
      throw new SyntaxError(`specified shape (${shape}) is compatible with input array or length (${array.length})`);
    return newArray;
  }
  static getInputShape(matrix = []) {
    if (Array.isArray(matrix) === false || !matrix[0] || !matrix[0].length || Array.isArray(matrix[0]) === false)
      throw new TypeError("input must be a matrix");
    const dim = [];
    const x_dimensions = matrix[0].length;
    let vectors = matrix;
    matrix.forEach((vector) => {
      if (vector.length !== x_dimensions)
        throw new SyntaxError("input must have the same length in each row");
    });
    for (; ; ) {
      dim.push(vectors.length);
      if (Array.isArray(vectors[0])) {
        vectors = vectors[0];
      } else {
        break;
      }
    }
    return dim;
  }
  exportConfiguration() {
    return {
      type: this.type,
      settings: this.settings,
      trained: this.trained,
      compiled: this.compiled,
      xShape: this.xShape,
      yShape: this.yShape,
      layers: this.layers
    };
  }
  importConfiguration(configuration) {
    this.type = configuration.type || this.type;
    this.settings = {
      ...this.settings,
      ...configuration.settings
    };
    this.trained = configuration.trained || this.trained;
    this.compiled = configuration.compiled || this.compiled;
    this.xShape = configuration.xShape || this.xShape;
    this.yShape = configuration.yShape || this.yShape;
    this.layers = configuration.layers || this.layers;
  }
  train(x_matrix, y_matrix) {
    throw new ReferenceError("train method is not implemented");
  }
  calculate(matrix) {
    throw new ReferenceError("calculate method is not implemented");
  }
  async loadModel(options) {
    this.model = await this.tf.loadLayersModel(options);
    this.xShape = this.model.inputs[0].shape;
    this.yShape = this.model.outputs[0].shape;
    this.trained = true;
    return this.model;
  }
  async saveModel(options) {
    const savedStatus = await this.model.save(options);
    return savedStatus;
  }
  async predict(input_matrix, options) {
    if (!input_matrix || Array.isArray(input_matrix) === false)
      throw new Error("invalid input matrix");
    const config = {
      json: true,
      probability: true,
      ...options
    };
    const x_matrix = Array.isArray(input_matrix) || config.skip_matrix_check ? input_matrix : [
      input_matrix
    ];
    return this.calculate(x_matrix).data().then((predictions) => {
      if (config.json === false) {
        return predictions;
      } else {
        if (!this.yShape)
          throw new Error("Model is missing yShape");
        const shape = [x_matrix.length, this.yShape[1]];
        const predictionValues = config.probability === false ? Array.from(predictions).map(Math.round) : Array.from(predictions);
        return this.reshape(predictionValues, shape);
      }
    }).catch((e) => {
      throw e;
    });
  }
};
function size(x) {
  let s = [];
  while (Array.isArray(x)) {
    s.push(x.length);
    x = x[0];
  }
  return s;
}
function _reshape(array, sizes) {
  var tmpArray = array;
  var tmpArray2;
  for (var sizeIndex = sizes.length - 1; sizeIndex > 0; sizeIndex--) {
    var size2 = sizes[sizeIndex];
    tmpArray2 = [];
    var length = tmpArray.length / size2;
    for (var i = 0; i < length; i++) {
      tmpArray2.push(tmpArray.slice(i * size2, (i + 1) * size2));
    }
    tmpArray = tmpArray2;
  }
  return tmpArray;
}
var DimensionError = class extends RangeError {
  actual;
  expected;
  relation;
  isDimensionError;
  constructor(actual, expected, relation) {
    const message = "Dimension mismatch (" + (Array.isArray(actual) ? "[" + actual.join(", ") + "]" : actual) + " != " + (Array.isArray(expected) ? "[" + expected.join(", ") + "]" : expected) + ")";
    super(message);
    this.actual = actual;
    this.expected = expected;
    this.relation = relation;
    this.message = message;
    this.name = "DimensionError";
    this.isDimensionError = true;
  }
};
function flatten(array) {
  if (!Array.isArray(array)) {
    return array;
  }
  let flat = [];
  array.forEach(function callback(value) {
    if (Array.isArray(value)) {
      value.forEach(callback);
    } else {
      flat.push(value);
    }
  });
  return flat;
}
async function asyncForEach(array, callback) {
  for (let index = 0; index < array.length; index++) {
    await callback(array[index], index, array);
  }
}

// src/base_neural_network.ts
var BaseNeuralNetwork = class extends TensorScriptModelInterface {
  constructor(options = {}, properties) {
    const config = Object.assign({
      layers: [],
      compile: {
        loss: "meanSquaredError",
        optimizer: "adam"
      },
      fit: {
        epochs: 100,
        batchSize: 5,
        verbose: 0
      }
    }, options);
    super(config, properties);
    this.type = "BaseNeuralNetwork";
    return this;
  }
  generateLayers(x_matrix, layers2) {
    throw new ReferenceError("generateLayers method is not implemented");
  }
  async train(x_matrix, y_matrix, layers2, x_test, y_test) {
    const xShape = this.getInputShape(x_matrix);
    const yShape = this.getInputShape(y_matrix);
    const xs = this.tf.tensor(x_matrix, xShape);
    const ys = this.tf.tensor(y_matrix, yShape);
    this.xShape = xShape;
    this.yShape = yShape;
    if (typeof this.compiled === "undefined" || this.compiled === false) {
      this.model = this.tf.sequential();
      this.generateLayers.call(this, x_matrix, y_matrix, layers2 || this.layers, x_test, y_test);
      this.model.compile(this.settings.compile);
      this.compiled = true;
    }
    await this.model.fit(xs, ys, this.settings.fit);
    this.trained = true;
    xs.dispose();
    ys.dispose();
    return this.model;
  }
  calculate(input_matrix, options) {
    if (!input_matrix || Array.isArray(input_matrix) === false)
      throw new Error("invalid input matrix");
    const predictionInput = Array.isArray(input_matrix[0]) ? input_matrix : [
      input_matrix
    ];
    const predictionTensor = this.tf.tensor(predictionInput);
    const prediction = this.model.predict(predictionTensor, options);
    predictionTensor.dispose();
    return prediction;
  }
};

// src/regression.ts
var DeepLearningRegression = class extends BaseNeuralNetwork {
  constructor(options = {}, properties) {
    const config = Object.assign({
      layers: [],
      layerPreference: "deep",
      compile: {
        loss: "meanSquaredError",
        optimizer: "adam"
      },
      fit: {
        epochs: 100,
        batchSize: 5
      }
    }, options);
    super(config, properties);
    this.type = "DeepLearningRegression";
    return this;
  }
  generateLayers(x_matrix, y_matrix, layers2) {
    const xShape = this.getInputShape(x_matrix);
    const yShape = this.getInputShape(y_matrix);
    const denseLayers = [];
    if (layers2) {
      denseLayers.push(...layers2);
    } else if (this.settings.layerPreference === "deep") {
      denseLayers.push({ units: xShape[1], inputShape: [xShape[1]], kernelInitializer: "randomNormal", activation: "relu" });
      denseLayers.push({ units: parseInt(String(Math.ceil(xShape[1] / 2)), 10), kernelInitializer: "randomNormal", activation: "relu" });
      denseLayers.push({ units: yShape[1], kernelInitializer: "randomNormal" });
    } else {
      denseLayers.push({ units: xShape[1] * 2, inputShape: [xShape[1]], kernelInitializer: "randomNormal", activation: "relu" });
      denseLayers.push({ units: yShape[1], kernelInitializer: "randomNormal" });
    }
    this.layers = denseLayers;
    denseLayers.forEach((layer) => {
      this.model.add(this.tf.layers.dense(layer));
    });
  }
};

// src/classification.ts
var DeepLearningClassification = class extends BaseNeuralNetwork {
  constructor(options = {}, properties) {
    const config = Object.assign({
      layers: [],
      compile: {
        loss: "categoricalCrossentropy",
        optimizer: "adam"
      },
      fit: {
        epochs: 100,
        batchSize: 5
      }
    }, options);
    super(config, properties);
    this.type = "DeepLearningClassification";
    return this;
  }
  generateLayers(x_matrix, y_matrix, layers2) {
    const xShape = this.getInputShape(x_matrix);
    const yShape = this.getInputShape(y_matrix);
    this.yShape = yShape;
    this.xShape = xShape;
    const denseLayers = [];
    if (layers2) {
      denseLayers.push(...layers2);
    } else {
      denseLayers.push({ units: xShape[1] * 2, inputDim: xShape[1], activation: "relu" });
      denseLayers.push({ units: yShape[1], activation: "softmax" });
    }
    this.layers = denseLayers;
    denseLayers.forEach((layer) => {
      this.model.add(this.tf.layers.dense(layer));
    });
  }
};

// src/logistic_regression.ts
var LogisticRegression = class extends BaseNeuralNetwork {
  constructor(options = {}, properties) {
    const config = Object.assign({
      layers: [],
      type: "simple",
      compile: {
        loss: "meanSquaredError",
        optimizer: "rmsprop"
      },
      fit: {
        epochs: 100,
        batchSize: 5
      }
    }, options);
    super(config, properties);
    this.type = "LogisticRegression";
    return this;
  }
  generateLayers(x_matrix, y_matrix, layers2, x_test, y_test) {
    const xShape = this.getInputShape(x_matrix);
    const yShape = this.getInputShape(y_matrix);
    this.yShape = yShape;
    this.xShape = xShape;
    const denseLayers = [];
    if (layers2) {
      denseLayers.push(...layers2);
    } else if (this.settings.type === "class" && this.settings.compile) {
      denseLayers.push({ units: 1, inputDim: xShape[1], activation: "sigmoid" });
      this.settings.compile.loss = "binaryCrossentropy";
    } else if (this.settings.type === "l1l2" && this.settings.compile) {
      const kernelRegularizer = this.tf.regularizers.l1l2({ l1: 0.01, l2: 0.01 });
      denseLayers.push({ units: 1, inputDim: xShape[1], activation: "sigmoid", kernelRegularizer });
      this.settings.compile.loss = "binaryCrossentropy";
    } else {
      denseLayers.push({ units: 1, inputShape: [xShape[1]] });
    }
    this.layers = denseLayers;
    denseLayers.forEach((layer) => {
      this.model.add(this.tf.layers.dense(layer));
    });
    if (x_test && y_test && this.settings && this.settings.fit) {
      this.settings.fit.validationData = [x_test, y_test];
    }
  }
};

// src/multiple_linear_regression.ts
var MultipleLinearRegression = class extends BaseNeuralNetwork {
  constructor(options = {}, properties = {}) {
    const config = Object.assign({
      layers: [],
      compile: {
        loss: "meanSquaredError",
        optimizer: "sgd"
      },
      fit: {
        epochs: 500,
        batchSize: 5
      }
    }, options);
    super(config, properties);
    this.type = "MultipleLinearRegression";
    return this;
  }
  generateLayers(x_matrix, y_matrix, layers2) {
    const xShape = this.getInputShape(x_matrix);
    const yShape = this.getInputShape(y_matrix);
    const denseLayers = [];
    if (layers2) {
      denseLayers.push(...layers2);
    } else {
      denseLayers.push({ units: yShape[1], inputShape: [xShape[1]] });
    }
    this.layers = denseLayers;
    denseLayers.forEach((layer) => {
      this.model.add(this.tf.layers.dense(layer));
    });
  }
};

// src/lstm_time_series.ts
var import_lodash = __toESM(require("lodash.range"), 1);
var LSTMTimeSeries = class extends BaseNeuralNetwork {
  static createDataset(dataset = [], look_back = 1) {
    const dataX = new Array();
    const dataY = new Array();
    for (let index in (0, import_lodash.default)(dataset.length - look_back - 1)) {
      let i = parseInt(index);
      let a = dataset.slice(i, i + look_back);
      dataX.push(a);
      dataY.push(dataset[i + look_back]);
    }
    return [dataX, dataY];
  }
  static getTimeseriesShape(x_timeseries) {
    const time_steps = this.settings.timeSteps;
    const xShape = this.getInputShape(x_timeseries);
    const _samples = xShape[0];
    const _timeSteps = time_steps;
    const _features = xShape[1];
    const newShape = this.settings.mulitpleTimeSteps || this.settings.stateful ? [_samples, _features, _timeSteps] : [_samples, _timeSteps, _features];
    return newShape;
  }
  static getTimeseriesDataSet(timeseries, look_back) {
    const lookback = look_back || this.settings.lookback;
    const matrices = LSTMTimeSeries.createDataset.call(this, timeseries, lookback);
    const x_matrix = matrices[0];
    const y_matrix = matrices[1];
    const x_matrix_timeseries = BaseNeuralNetwork.reshape(x_matrix, [x_matrix.length, lookback]);
    const xShape = BaseNeuralNetwork.getInputShape(x_matrix_timeseries);
    const yShape = BaseNeuralNetwork.getInputShape(y_matrix);
    return {
      yShape,
      xShape,
      y_matrix,
      x_matrix: x_matrix_timeseries
    };
  }
  createDataset;
  getTimeseriesShape;
  getTimeseriesDataSet;
  constructor(options = {}, properties) {
    const config = Object.assign({
      layers: [],
      type: "simple",
      stateful: false,
      stacked: false,
      mulitpleTimeSteps: false,
      lookback: 1,
      batchSize: 1,
      timeSteps: 1,
      learningRate: 0.1,
      compile: {
        loss: "meanSquaredError",
        optimizer: "adam"
      },
      fit: {
        epochs: 100,
        batchSize: 1
      }
    }, options);
    super(config, properties);
    this.type = "LSTMTimeSeries";
    this.createDataset = LSTMTimeSeries.createDataset;
    this.getTimeseriesDataSet = LSTMTimeSeries.getTimeseriesDataSet;
    this.getTimeseriesShape = LSTMTimeSeries.getTimeseriesShape;
    return this;
  }
  generateLayers(x_matrix, y_matrix, layers2) {
    const xShape = this.getInputShape(x_matrix);
    const yShape = this.getInputShape(y_matrix);
    this.yShape = yShape;
    this.xShape = xShape;
    const lstmLayers = [];
    const rnnLayers = [];
    const denseLayers = [];
    if (layers2) {
      if (layers2.lstmLayers && layers2.lstmLayers.length)
        lstmLayers.push(...layers2.lstmLayers);
      if (layers2.rnnLayers && layers2.rnnLayers.length)
        rnnLayers.push(...layers2.rnnLayers);
      if (layers2.denseLayers && layers2.denseLayers.length)
        denseLayers.push(...layers2.denseLayers);
    } else if (this.settings && this.settings.fit && this.settings.stateful) {
      const batchInputShape = [this.settings.fit.batchSize, this.settings.lookback, 1];
      rnnLayers.push({ units: 4, batchInputShape, returnSequences: true });
      rnnLayers.push({ units: 4, batchInputShape });
      denseLayers.push({ units: yShape[1] });
    } else {
      const inputShape = [1, this.settings.lookback];
      lstmLayers.push({ units: 4, inputShape });
      denseLayers.push({ units: yShape[1] });
    }
    if (lstmLayers.length) {
      lstmLayers.forEach((layer) => {
        this.model.add(this.tf.layers.lstm(layer));
      });
    }
    if (rnnLayers.length) {
      rnnLayers.forEach((layer) => {
        this.model.add(this.tf.layers.simpleRNN(layer));
      });
    }
    if (denseLayers.length) {
      denseLayers.forEach((layer) => {
        this.model.add(this.tf.layers.dense(layer));
      });
    }
    this.layers = {
      lstmLayers,
      rnnLayers,
      denseLayers
    };
  }
  async train(x_timeseries, y_timeseries, layers2, x_test, y_test) {
    let yShape;
    let x_matrix;
    let y_matrix;
    const look_back = this.settings.lookback;
    if (y_timeseries) {
      x_matrix = x_timeseries;
      y_matrix = y_timeseries;
    } else {
      const matrices = this.createDataset(x_timeseries, look_back);
      x_matrix = matrices[0];
      y_matrix = matrices[1];
      yShape = this.getInputShape(y_matrix);
    }
    const timeseriesShape = this.getTimeseriesShape(x_matrix);
    const x_matrix_timeseries = BaseNeuralNetwork.reshape(x_matrix, timeseriesShape);
    const xs = this.tf.tensor(x_matrix_timeseries, timeseriesShape);
    const ys = this.tf.tensor(y_matrix, yShape);
    this.xShape = timeseriesShape;
    this.yShape = yShape;
    if (this.compiled === false) {
      this.model = this.tf.sequential();
      this.generateLayers.call(
        this,
        x_matrix_timeseries,
        y_matrix,
        layers2 || this.layers
      );
      this.model.compile(this.settings.compile);
      if (this.settings.fit && this.settings.stateful) {
        this.settings.fit.shuffle = false;
      }
      this.compiled = true;
    }
    await this.model.fit(xs, ys, this.settings.fit);
    this.trained = true;
    xs.dispose();
    ys.dispose();
    return this.model;
  }
  calculate(x_matrix) {
    const timeseriesShape = this.getTimeseriesShape(x_matrix);
    const input_matrix = BaseNeuralNetwork.reshape(x_matrix, timeseriesShape);
    return super.calculate(input_matrix);
  }
  async predict(input_matrix, options = {}) {
    if (this.settings.stateful && input_matrix.length > 1) {
      return Promise.all(input_matrix.map((input) => super.predict([input], options)));
    } else {
      return super.predict.call(this, input_matrix, options);
    }
  }
};

// src/lstm_multivariate_time_series.ts
var import_lodash2 = __toESM(require("lodash.range"), 1);
var LSTMMultivariateTimeSeries = class extends LSTMTimeSeries {
  static createDataset(dataset = [], look_back = 1) {
    const features = this.settings && this.settings.features ? this.settings.features : dataset[0].length - 1;
    const n_in = look_back || this.settings.lookback || 1;
    const n_out = this.settings && this.settings.outputs ? this.settings.outputs : 1;
    const series = LSTMMultivariateTimeSeries.seriesToSupervised(dataset, n_in, n_out);
    const dropped = LSTMMultivariateTimeSeries.getDropableColumns(features, n_in, n_out);
    const droppedColumns = LSTMMultivariateTimeSeries.drop(series, dropped);
    const dropped_c_columns = [0];
    const original_dropped_c_columns = [0, droppedColumns[0].length - 1];
    for (let i = 0; i < n_in; i++) {
      dropped_c_columns.push((i + 1) * features + 1);
    }
    const y = pivotVector(droppedColumns)[droppedColumns[0].length - 1];
    const x = LSTMMultivariateTimeSeries.drop(
      droppedColumns,
      original_dropped_c_columns
    );
    return [x, y, features];
  }
  static drop(data, columns) {
    return data.reduce((cols, row, i) => {
      cols[i] = [];
      row.forEach((col, idx) => {
        if (columns.indexOf(idx) === -1) {
          cols[i].push(col);
        }
      });
      return cols;
    }, []);
  }
  static seriesToSupervised(data, n_in = 1, n_out = 1) {
    if (n_out !== 1)
      throw new RangeError("Only 1 future iteration supported");
    if (data && Array.isArray(data) && Array.isArray(data[0]) && data[0].length < 2)
      throw new RangeError("Must include at least two columns, the first column must be the dependent variable, followed by independent variables");
    let cols = [];
    for (let x in data) {
      x = Number(parseInt(x));
      let maxIndex = x + n_in + n_out;
      if (maxIndex > data.length)
        break;
      cols[x] = [];
      for (let i in (0, import_lodash2.default)(n_in)) {
        i = parseInt(i);
        cols[x].push(...data[x + i]);
      }
      for (let j in (0, import_lodash2.default)(n_out)) {
        j = parseInt(j);
        cols[x].push(...data[x + j + n_in]);
      }
    }
    return cols;
  }
  static getDropableColumns(features, n_in, n_out) {
    if (n_out !== 1)
      throw new RangeError("Only 1 future iteration supported");
    const cols = features + 1;
    const total_cols = cols * n_in + cols * n_out;
    return (0, import_lodash2.default)(total_cols - cols + 1, total_cols);
  }
  static getTimeseriesShape(x_timeseries) {
    const time_steps = this.settings && this.settings.lookback ? this.settings.lookback : 1;
    const xShape = this.getInputShape(x_timeseries);
    const _samples = xShape[0];
    const _timeSteps = time_steps;
    const _features = this.settings && this.settings.features ? this.settings.features : xShape[1];
    const newShape = [_samples, _timeSteps, _features];
    return newShape;
  }
  static getTimeseriesDataSet(timeseries, look_back) {
    const lookBack = look_back || this.settings.lookback;
    const matrices = LSTMMultivariateTimeSeries.createDataset.call(this, timeseries, lookBack);
    const x_matrix = matrices[0];
    const y_matrix_m = LSTMMultivariateTimeSeries.reshape(matrices[1], [matrices[1].length, 1]);
    const timeseriesShape = LSTMMultivariateTimeSeries.getTimeseriesShape.call(this, x_matrix);
    const x_matrix_timeseries = LSTMMultivariateTimeSeries.reshape(x_matrix, timeseriesShape);
    const xShape = LSTMMultivariateTimeSeries.getInputShape(x_matrix_timeseries);
    const y_matrix = pivotVector(y_matrix_m)[y_matrix_m[0].length - 1];
    const yShape = LSTMMultivariateTimeSeries.getInputShape(y_matrix_m);
    return {
      yShape,
      xShape,
      y_matrix,
      x_matrix: x_matrix_timeseries
    };
  }
  createDataset;
  seriesToSupervised;
  getDropableColumns;
  drop;
  getTimeseriesShape;
  getTimeseriesDataSet;
  constructor(options = {}, properties) {
    const config = Object.assign({
      layers: [],
      stateful: true,
      mulitpleTimeSteps: false,
      lookback: 1,
      features: void 0,
      outputs: 1,
      learningRate: 0.1,
      compile: {
        loss: "meanSquaredError",
        optimizer: "adam"
      },
      fit: {
        epochs: 100,
        batchSize: 1,
        shuffle: false
      }
    }, options);
    super(config, properties);
    this.type = "LSTMMultivariateTimeSeries";
    this.createDataset = LSTMMultivariateTimeSeries.createDataset;
    this.seriesToSupervised = LSTMMultivariateTimeSeries.seriesToSupervised;
    this.getDropableColumns = LSTMMultivariateTimeSeries.getDropableColumns;
    this.drop = LSTMMultivariateTimeSeries.drop;
    this.getTimeseriesShape = LSTMMultivariateTimeSeries.getTimeseriesShape;
    this.getTimeseriesDataSet = LSTMMultivariateTimeSeries.getTimeseriesDataSet;
    return this;
  }
  generateLayers(x_matrix, y_matrix, layers2) {
    const xShape = this.getInputShape(x_matrix);
    const yShape = this.getInputShape(y_matrix);
    this.yShape = yShape;
    this.xShape = xShape;
    const lstmLayers = [];
    const denseLayers = [];
    if (layers2) {
      if (layers2.lstmLayers && layers2.lstmLayers.length)
        lstmLayers.push(...layers2.lstmLayers);
      if (layers2.denseLayers && layers2.denseLayers.length)
        denseLayers.push(...layers2.denseLayers);
    } else {
      const inputShape = [xShape[1], xShape[2]];
      lstmLayers.push({ units: 10, inputShape });
      denseLayers.push({ units: yShape[1] });
    }
    if (lstmLayers.length) {
      lstmLayers.forEach((layer) => {
        this.model.add(this.tf.layers.lstm(layer));
      });
    }
    if (denseLayers.length) {
      denseLayers.forEach((layer) => {
        this.model.add(this.tf.layers.dense(layer));
      });
    }
    this.layers = {
      lstmLayers,
      denseLayers
    };
  }
  async train(x_timeseries, y_timeseries, layers2, x_test, y_test) {
    let xShape;
    let yShape;
    let x_matrix;
    let y_matrix;
    const look_back = this.settings.lookback;
    if (y_timeseries) {
      x_matrix = x_timeseries;
      y_matrix = y_timeseries;
      xShape = this.getInputShape(x_matrix);
      yShape = this.getInputShape(y_matrix);
    } else {
      const matrices = this.createDataset(x_timeseries, look_back);
      x_matrix = matrices[0];
      y_matrix = this.reshape(matrices[1], [matrices[1].length, 1]);
      xShape = this.getInputShape(x_matrix);
      yShape = this.getInputShape(y_matrix);
    }
    const timeseriesShape = this.getTimeseriesShape(x_matrix);
    const x_matrix_timeseries = LSTMMultivariateTimeSeries.reshape(x_matrix, timeseriesShape);
    const xs = this.tf.tensor(x_matrix_timeseries, timeseriesShape);
    const ys = this.tf.tensor(y_matrix, yShape);
    this.xShape = timeseriesShape;
    this.yShape = yShape;
    if (this.compiled === false) {
      this.model = this.tf.sequential();
      this.generateLayers.call(
        this,
        x_matrix_timeseries,
        y_matrix,
        layers2 || this.layers
      );
      this.model.compile(this.settings.compile);
      this.compiled = true;
    }
    if (x_test && y_test && this.settings && this.settings.fit) {
      this.settings.fit.validation_data = [x_test, y_test];
    }
    await this.model.fit(xs, ys, this.settings.fit);
    this.trained = true;
    xs.dispose();
    ys.dispose();
    return this.model;
  }
};
function pivotVector(vectors = []) {
  return vectors.reduce((result, val, index) => {
    val.forEach((vecVal, i) => {
      index === 0 ? result.push([vecVal]) : result[i].push(vecVal);
    });
    return result;
  }, []);
}

// src/text_embedding.ts
var import_universal_sentence_encoder = require("@tensorflow-models/universal-sentence-encoder/dist/universal-sentence-encoder.js");
var import_axios = __toESM(require("axios"), 1);
var BASE_PATH = "https://storage.googleapis.com/tfjs-models/savedmodel/universal_sentence_encoder";
async function loadVocabulary(pathToVocabulary) {
  const vocabulary = await import_axios.default.get(pathToVocabulary);
  return vocabulary.data;
}
async function loadTokenizer() {
  const vocabulary = await loadVocabulary(`${BASE_PATH}/vocab.json`);
  const tokenizer2 = new import_universal_sentence_encoder.Tokenizer(vocabulary);
  return tokenizer2;
}
var model;
var tokenizer;
var TextEmbedding = class extends TensorScriptModelInterface {
  constructor(options = {}, properties) {
    const config = Object.assign({}, options);
    super(config, properties);
    this.type = "TextEmbedding";
    return this;
  }
  async train() {
    const promises = [];
    if (!model)
      promises.push((0, import_universal_sentence_encoder.load)());
    else
      promises.push(Promise.resolve(model));
    if (!tokenizer)
      promises.push(loadTokenizer());
    else
      promises.push(Promise.resolve(tokenizer));
    const USE = await Promise.all(promises);
    if (!model)
      model = USE[0];
    if (!tokenizer)
      tokenizer = USE[1];
    this.model = model;
    this.tokenizer = tokenizer;
    this.trained = true;
    this.compiled = true;
    return this.model;
  }
  calculate(input_array, options = {}) {
    if (!input_array || Array.isArray(input_array) === false)
      throw new Error("invalid input array of sentences");
    const embeddings = this.model.embed(input_array);
    return embeddings;
  }
  async predict(input_array, options = {}) {
    const config = Object.assign({
      json: true,
      probability: true
    }, options);
    const embeddings = await this.calculate(input_array, options);
    const predictions = await embeddings.data();
    if (config.json === false) {
      return predictions;
    } else {
      const shape = [input_array.length, 512];
      const predictionValues = options.probability === false ? Array.from(predictions).map(Math.round) : Array.from(predictions);
      return this.reshape(predictionValues, shape);
    }
  }
};

// src/feature_embedding.ts
var import_lodash3 = __toESM(require("lodash.range"), 1);
var import_tsne_js = __toESM(require("tsne-js"), 1);
var FeatureEmbedding = class extends BaseNeuralNetwork {
  featureToId;
  idToFeature;
  featureIds;
  numberOfFeatures;
  importedEmbeddings;
  static async getFeatureDataSet({ inputMatrixFeatures, PAD = "PAD", initialIdToFeature, initialFeatureToId }) {
    let featIndex = initialFeatureToId ? Object.keys(initialFeatureToId).length : 1;
    const idToFeature = {
      0: this && this.settings && this.settings.PAD ? this.settings.PAD : PAD,
      ...initialIdToFeature
    };
    const featureToId = inputMatrixFeatures.reduce((result, inputFeatureArray) => {
      inputFeatureArray.forEach((inputFeature) => {
        if (!result[inputFeature]) {
          if (idToFeature[featIndex])
            featIndex++;
          result[inputFeature] = featIndex;
          idToFeature[featIndex] = inputFeature;
          featIndex++;
        }
      });
      return result;
    }, {
      [PAD]: 0,
      ...initialFeatureToId
    });
    const featureIds = inputMatrixFeatures.map(
      (inputFeatureArray) => inputFeatureArray.map((inputFeature) => featureToId[inputFeature])
    );
    const numberOfFeatures = Object.keys(featureToId).length;
    return {
      featureToId,
      idToFeature,
      featureIds,
      numberOfFeatures
    };
  }
  static getMergedArray(base = [], merger = [], append = false, truncate = true) {
    let arr = new Array().concat(base);
    if (append)
      arr.splice(base.length - merger.length, merger.length, ...merger);
    else
      arr.splice(0, merger.length, ...merger);
    if (truncate && append)
      return arr.slice(-1 * base.length);
    else if (truncate)
      return arr.slice(0, base.length);
    else
      return arr;
  }
  static async getContextPairs({ inputMatrix, numberOfFeatures, window_size = 2, tf: tf2 }) {
    const tensorflow = this && this.tf ? this.tf : getBackend();
    const context_length = (this && this.settings && this.settings.windowSize ? this.settings.windowSize : window_size) * 2;
    const [emptyXVector, emptyYVector] = await Promise.all([
      tensorflow.zeros([context_length]).array(),
      tensorflow.zeros([numberOfFeatures]).array()
    ]);
    const x = [];
    const y = [];
    inputMatrix.forEach((inputVector, inputVectorIndex) => {
      inputVector.forEach((word, index) => {
        if (word != 0) {
          const output = new Array().concat(emptyYVector);
          const inputMerger = new Array().concat(inputMatrix[inputVectorIndex]);
          inputMerger.splice(index, 1);
          const input = FeatureEmbedding.getMergedArray(emptyXVector, inputMerger, true);
          output[word] = 1;
          x.push(input);
          y.push(output);
        }
      });
    });
    return {
      context_length,
      emptyXVector,
      emptyYVector,
      x,
      y
    };
  }
  getMergedArray;
  getFeatureDataSet;
  getContextPairs;
  constructor(options = {}, properties) {
    const config = {
      layers: [],
      type: "cbow",
      compile: {
        loss: "categoricalCrossentropy",
        optimizer: "rmsprop"
      },
      fit: {
        epochs: 15,
        batchSize: 1
      },
      embedSize: 50,
      windowSize: 2,
      PAD: "PAD",
      streamInputMatrix: true,
      initialLayerInitializerType: "randomNormal",
      initialLayerInitializerOptions: { seed: 1 },
      ...options
    };
    super(config, properties);
    this.type = "FeatureEmbedding";
    this.featureToId;
    this.idToFeature;
    this.featureIds;
    this.numberOfFeatures;
    this.getMergedArray = FeatureEmbedding.getMergedArray;
    this.getFeatureDataSet = FeatureEmbedding.getFeatureDataSet.bind(this);
    this.getContextPairs = FeatureEmbedding.getContextPairs.bind(this);
    return this;
  }
  generateLayers(x_matrix, y_matrix, layers2) {
    if (!this.numberOfFeatures)
      throw ReferenceError(`${this.settings.name} model is missing numberOfFeatures`);
    if (!this.settings.embedSize)
      throw ReferenceError(`${this.settings.name} model is missing embedSize`);
    const yShape = [this.numberOfFeatures, this.settings.embedSize];
    this.yShape = yShape;
    const denseLayers = [];
    denseLayers.push({
      units: this.numberOfFeatures,
      inputDim: this.numberOfFeatures,
      outputDim: this.settings.embedSize,
      inputLength: (this.settings.windowSize || 2) * 2,
      embeddingsInitializer: this.settings.initialLayerInitializerType ? this.tf.initializers[this.settings.initialLayerInitializerType](this.settings.initialLayerInitializerOptions) : void 0
    });
    denseLayers.push({
      lambdaFunction: "result = tf.mean(input,1,true)",
      lambdaOutputShape: [this.numberOfFeatures, this.settings.embedSize]
    });
    denseLayers.push({ units: this.numberOfFeatures, activation: "softmax" });
    this.layers = denseLayers;
    this.model.add(this.tf.layers.embedding(denseLayers[0]));
    this.model.add(this.tf.layers.flatten());
    this.model.add(this.tf.layers.dense(denseLayers[2]));
    if (layers2 && layers2.length && layers2[0].weights) {
      const originalModelWeights = this.model.getWeights();
      originalModelWeights[0] = layers2[0].weights;
      this.model.setWeights(originalModelWeights);
    }
  }
  async trainOnBatch({ x_input_matrix, y_output_matrix, epoch, trainingLoss, inputVectorIndex, inputVectorLength }) {
    let loss = Infinity;
    if (this.settings.fit?.callbacks?.onEpochBegin)
      this.settings.fit?.callbacks?.onEpochBegin(epoch, { loss: trainingLoss });
    await asyncForEach(x_input_matrix, async (x_input, xIndex) => {
      if (this.settings.fit?.callbacks?.onBatchBegin)
        this.settings.fit?.callbacks?.onBatchBegin(xIndex, { loss: trainingLoss });
      const y_output = y_output_matrix[xIndex];
      const xShape = this.getInputShape([x_input]);
      const xs = this.tf.tensor(x_input, xShape);
      const yShape = this.getInputShape([y_output]);
      const ys = this.tf.tensor(y_output, yShape);
      loss = await this.model.trainOnBatch(xs, ys);
      if (this.settings.fit?.callbacks?.onYield)
        this.settings.fit?.callbacks?.onYield(epoch, xIndex, {
          loss,
          inputVectorIndex,
          inputVectorLength,
          completion: `${(100 * (((inputVectorIndex || xIndex) + 1) / (inputVectorLength || x_input_matrix.length))).toFixed(2)}%`
        });
      if (this.settings.fit?.callbacks?.onBatchEnd)
        this.settings.fit?.callbacks?.onBatchEnd(xIndex, { loss });
      xs.dispose();
      ys.dispose();
    });
    if (this.settings.fit?.callbacks?.onEpochEnd)
      this.settings.fit?.callbacks?.onEpochEnd(epoch, { loss });
    return {
      loss
    };
  }
  async generateBatch({ epoch }) {
    if (!this.featureIds)
      throw ReferenceError(`${this.settings.name} model is missing featureIds`);
    const preTransformedMatrix = this.featureIds;
    const context_length = (this && this.settings && this.settings.windowSize ? this.settings.windowSize : 2) * 2;
    const [emptyXVector, emptyYVector] = await Promise.all([
      this.tf.zeros([context_length]).array(),
      this.tf.zeros([this.numberOfFeatures]).array()
    ]);
    let x_input_matrix = [];
    let y_output_matrix = [];
    let trainingLoss = Infinity;
    await asyncForEach(preTransformedMatrix, async (inputVector, inputVectorIndex) => {
      await asyncForEach(inputVector, async (word, index) => {
        if (this.settings.checkInputMatrix && this.numberOfFeatures && word >= this.numberOfFeatures) {
          console.warn("invalid word in corpus", {
            trainingLoss,
            word,
            "this.idToFeature[word]": this.idToFeature && this.idToFeature[word],
            "this.numberOfFeatures": this.numberOfFeatures
          });
        } else if (word != 0) {
          const output = new Array().concat(emptyYVector);
          const inputMerger = new Array().concat(preTransformedMatrix[inputVectorIndex]);
          inputMerger.splice(index, 1);
          const input = FeatureEmbedding.getMergedArray(emptyXVector, inputMerger, true);
          output[word] = 1;
          x_input_matrix = [input];
          y_output_matrix = [output];
          if (this.settings.checkInputMatrix && this.numberOfFeatures && this.numberOfFeatures > 0 && input.filter((wordInput) => this.numberOfFeatures && wordInput >= this.numberOfFeatures).length) {
            console.warn("Input matrix contains unknown weight", { input, "this.numberOfFeatures": this.numberOfFeatures });
          } else {
            x_input_matrix = [input];
            y_output_matrix = [output];
            const modelStatus = await this.trainOnBatch({ x_input_matrix, y_output_matrix, epoch, trainingLoss, inputVectorIndex, inputVectorLength: preTransformedMatrix.length });
            trainingLoss = modelStatus.loss;
          }
        }
      });
    });
    return {
      loss: trainingLoss
    };
  }
  async exportEmbeddings() {
    if (this.trained !== true)
      throw new ReferenceError("The model has to be trained before embeddings can be exported");
    const weights = await this.predict();
    const labeledWeights = this.labelWeights(weights);
    return {
      featureToId: this.featureToId,
      idToFeature: this.idToFeature,
      featureIds: this.featureIds,
      numberOfFeatures: this.numberOfFeatures,
      labeledWeights
    };
  }
  async importEmbeddings({ featureToId, idToFeature, featureIds, numberOfFeatures, labeledWeights, addNewWeights = true, inputMatrixFeatures, fixImportedWeights = false }) {
    this.model = void 0;
    let updatedModelProperties;
    if (addNewWeights) {
      if (inputMatrixFeatures) {
        updatedModelProperties = await this.getFeatureDataSet({ inputMatrixFeatures, initialIdToFeature: idToFeature, initialFeatureToId: featureToId });
        featureToId = updatedModelProperties?.featureToId;
        idToFeature = updatedModelProperties?.idToFeature;
        featureIds = updatedModelProperties?.featureIds;
        numberOfFeatures = updatedModelProperties?.numberOfFeatures;
      }
      if (featureToId) {
        await asyncForEach(Object.keys(featureToId), async (weightLabel) => {
          if (!labeledWeights[weightLabel] || !labeledWeights[weightLabel].length) {
            labeledWeights[weightLabel] = await this.tf.randomUniform([1, this.settings.embedSize], -1, 1).array();
            if (Array.isArray(labeledWeights[weightLabel][0])) {
              if (typeof labeledWeights[weightLabel].flat === "function")
                labeledWeights[weightLabel] = labeledWeights[weightLabel].flat();
              else
                labeledWeights[weightLabel] = labeledWeights[weightLabel].reduce((acc, val) => acc.concat(val), []);
            }
          }
        });
      }
    }
    if (fixImportedWeights) {
      Object.keys(labeledWeights).forEach((weightLabel) => {
        if (Array.isArray(labeledWeights[weightLabel][0])) {
          if (typeof labeledWeights[weightLabel].flat === "function")
            labeledWeights[weightLabel] = labeledWeights[weightLabel].flat();
          else
            labeledWeights[weightLabel] = labeledWeights[weightLabel].reduce((acc, val) => acc.concat(val), []);
        }
      });
    }
    const firstLabeledWeight = Object.keys(labeledWeights)[0];
    if (!firstLabeledWeight || !labeledWeights[firstLabeledWeight] || labeledWeights[firstLabeledWeight].length !== this.settings.embedSize)
      throw new RangeError(`imported weights (${labeledWeights[firstLabeledWeight] ? labeledWeights[firstLabeledWeight].length : "firstLabeledWeight:undefined"}) must have the same embedding size as model (${this.settings.embedSize})`);
    const trainedWeights = this.tf.variable(this.tf.tensor(Object.values(labeledWeights)));
    this.featureToId = featureToId;
    this.idToFeature = idToFeature;
    this.featureIds = featureIds;
    this.numberOfFeatures = numberOfFeatures;
    if (trainedWeights.shape[0] !== this.numberOfFeatures) {
      console.warn("INVALID NUMBER OF this.numberOfFeatures", {
        "trainedWeights.shape[0]": trainedWeights.shape[0],
        "this.numberOfFeatures": this.numberOfFeatures
      });
      this.numberOfFeatures = trainedWeights.shape[0];
    }
    this.compileModel({ layers: [{ weights: trainedWeights }] });
    this.importedEmbeddings = true;
  }
  compileModel({ layers: layers2 } = {}) {
    this.model = void 0;
    this.model = this.tf.sequential();
    this.generateLayers.call(this, [], [], layers2 || this.layers);
    this.model.compile(this.settings.compile);
    this.compiled = true;
  }
  async train(x_matrix, y_matrix, layers2) {
    if (!this.featureToId || !this.idToFeature || !this.featureIds || !this.numberOfFeatures) {
      const featureEmbedDataSet = await this.getFeatureDataSet({ inputMatrixFeatures: x_matrix });
      this.featureToId = featureEmbedDataSet.featureToId;
      this.idToFeature = featureEmbedDataSet.idToFeature;
      this.featureIds = featureEmbedDataSet.featureIds;
      this.numberOfFeatures = featureEmbedDataSet.numberOfFeatures;
    } else if (this.importedEmbeddings && x_matrix && x_matrix.length)
      this.featureIds = x_matrix;
    if (this.compiled === false)
      this.compileModel({ layers: layers2 });
    let loss = Infinity;
    if (this.settings.fit?.callbacks?.onTrainBegin)
      this.settings.fit?.callbacks?.onTrainBegin({ loss });
    await asyncForEach((0, import_lodash3.default)(0, this.settings.fit?.epochs), async (epoch) => {
      if (this.settings.streamInputMatrix) {
        let modelStatus = await this.generateBatch({ epoch });
        loss = modelStatus.loss;
      } else {
        if (!this.numberOfFeatures)
          throw ReferenceError(`${this.settings.name} model is missing numberOfFeatures`);
        if (!this.featureIds)
          throw ReferenceError(`${this.settings.name} model is missing featureIds`);
        const cxt = await this.getContextPairs({ tf: this.tf, numberOfFeatures: this.numberOfFeatures, inputMatrix: this.featureIds });
        const x_input_matrix = cxt.x;
        const y_output_matrix = cxt.y;
        let modelStatus = await this.trainOnBatch({ x_input_matrix, y_output_matrix, epoch, trainingLoss: loss });
        loss = modelStatus.loss;
      }
    });
    if (this.settings.fit?.callbacks?.onTrainEnd)
      this.settings.fit?.callbacks?.onTrainEnd({ loss });
    this.loss = loss;
    this.trained = true;
    return this.model;
  }
  async calculate() {
    return this.model.getWeights()[0];
  }
  async predict(options = {}) {
    const predictions = await this.calculate();
    if (options.json === false) {
      return await predictions.data();
    } else {
      const arr = await predictions.array();
      if (!this.yShape)
        throw new Error("Model is missing yShape");
      return this.reshape(arr, predictions.shape);
    }
  }
  labelWeights(weights) {
    return weights.reduce((result, weight, index) => {
      if (this.idToFeature)
        result[this.idToFeature[index]] = weight;
      return result;
    }, {});
  }
  async reduceWeights(weights, options) {
    let model2 = new import_tsne_js.default({
      dim: 2,
      perplexity: 30,
      earlyExaggeration: 4,
      learningRate: 100,
      nIter: 1e3,
      metric: "euclidean",
      ...options
    });
    model2.init({
      data: weights,
      type: options ? options.type : "dense"
    });
    let [error, iter] = model2.run();
    let output = model2.getOutput();
    return output;
  }
  async findSimilarFeatures(weights, options = {}) {
    const tf2 = this.tf;
    const { features = [], limit = 5, labeledWeights, metric = "distance" } = options;
    const labeledFeatureWeights = labeledWeights || this.labelWeights(weights);
    if (this.settings && this.settings.PAD)
      delete labeledFeatureWeights[this.settings.PAD];
    return features.reduce((result, feature) => {
      const featureWeights = labeledFeatureWeights[feature];
      if (!featureWeights)
        throw new ReferenceError(`Invalid feature: ${feature}`);
      const sims = Object.keys(labeledFeatureWeights).map((searchFeature) => {
        const prox = tf2.tidy(() => {
          const proximity = tf2.metrics.cosineProximity(
            tf2.tensor(featureWeights),
            tf2.tensor(labeledFeatureWeights[searchFeature])
          );
          const distance = tf2.metrics.meanSquaredError(
            tf2.tensor(featureWeights),
            tf2.tensor(labeledFeatureWeights[searchFeature])
          );
          return [
            proximity.asScalar().dataSync()[0],
            distance.asScalar().dataSync()[0]
          ];
        });
        return {
          comparedFeature: searchFeature,
          proximity: prox[0],
          distance: prox[1]
        };
      }).sort((a, b) => metric === "distance" ? a.distance - b.distance : a.proximity - b.proximity);
      result[feature] = sims.slice(1, limit + 1);
      return result;
    }, {});
  }
};

// src/scikitjs_singleton.ts
var scikit = null;
function setScikit(scikitInput) {
  scikit = scikitInput;
}
function getScikit() {
  if (scikit === null) {
    throw Error(`
============================
Howdy \u{1F44B}\u{1F44B}. Looks like you are running @jsonstack/model but you haven't set a ScikitJS backend. 
To do so, simply import (or require) your scikit library, and call setScikit like so,

import * as tf from '@tensorflow/tfjs';
import * as scikit from 'scikitjs';
import * as jsm from '@jsonstack/model';
jsm.setBackend(tf);
jsm.setScikit(scikit);

That will let @jsonstack/model know you wish to use a scikitjs library to perform your calculations.
============================
    `);
  }
  return scikit;
}

// src/ml_model_interface.ts
var MachineLearningModelInterface = class extends TensorScriptModelInterface {
  scikit;
  constructor(options = {}, properties = {}) {
    let scikit2 = getScikit();
    const config = {
      ...options
    };
    super(config, properties);
    this.scikit = properties.scikit || scikit2;
    this.type = "MachineLearningModelInterface";
    return this;
  }
  explain() {
    return {
      coefficients: this.model.coef,
      intercept: this.model.intercept
    };
  }
};

// src/ml_regression.ts
var MachineLearningLinearRegression = class extends MachineLearningModelInterface {
  constructor(options = {}, properties) {
    const config = {
      layers: [],
      layerPreference: "deep",
      compile: {},
      fit: {
        callbacks: []
      },
      ...options
    };
    super(config, properties);
    if (options?.fit?.callbacks)
      config.fit.callbacks = [
        this.tf.callbacks.earlyStopping({ monitor: "mse", patience: 30 })
      ].concat(createModelFitCallback(options?.fit?.callbacks));
    this.type = "MachineLearningLinearRegression";
    this.model = new this.scikit.LinearRegression({
      modelFitOptions: config.fit
    });
    return this;
  }
  async train(x_matrix, y_matrix) {
    const xShape = this.getInputShape(x_matrix);
    const yShape = this.getInputShape(y_matrix);
    this.xShape = xShape;
    this.yShape = yShape;
    await this.model.fit(x_matrix, y_matrix, this.settings.fit);
    this.trained = true;
    this.compiled = true;
    return this.model;
  }
  calculate(input_matrix, options) {
    if (!input_matrix || Array.isArray(input_matrix) === false)
      throw new Error("invalid input matrix");
    const predictionInput = Array.isArray(input_matrix[0]) ? input_matrix : [
      input_matrix
    ];
    const predictionTensor = this.tf.tensor(predictionInput);
    const prediction = this.model.predict(predictionTensor, options);
    predictionTensor.dispose();
    return prediction;
  }
};

// src/ml_logistic_regression.ts
var MachineLearningLogisticRegression = class extends MachineLearningModelInterface {
  constructor(options = {}, properties) {
    const config = {
      layers: [],
      layerPreference: "deep",
      compile: {},
      fit: {
        callbacks: []
      },
      ...options
    };
    super(config, properties);
    if (options?.fit?.callbacks)
      config.fit.callbacks = [
        this.tf.callbacks.earlyStopping({ monitor: "loss", patience: 50 })
      ].concat(createModelFitCallback(options?.fit?.callbacks));
    this.type = "MachineLearningLogisticRegression";
    this.model = new this.scikit.LogisticRegression({
      modelFitOptions: config.fit
    });
    return this;
  }
  async train(x_matrix, y_matrix) {
    const trainable_y_matrix = y_matrix.map((y) => y[0]);
    const xShape = this.getInputShape(x_matrix);
    const yShape = this.getInputShape(y_matrix);
    this.xShape = xShape;
    this.yShape = yShape;
    await this.model.fit(x_matrix, trainable_y_matrix, this.settings.fit);
    this.trained = true;
    this.compiled = true;
    return this.model;
  }
  calculate(input_matrix, options) {
    if (!input_matrix || Array.isArray(input_matrix) === false)
      throw new Error("invalid input matrix");
    const predictionInput = Array.isArray(input_matrix[0]) ? input_matrix : [
      input_matrix
    ];
    const predictionTensor = this.tf.tensor(predictionInput);
    const prediction = this.model.predict(predictionTensor, options);
    predictionTensor.dispose();
    return prediction;
  }
};
// Annotate the CommonJS export names for ESM import in node:
0 && (module.exports = {
  BaseNeuralNetwork,
  DeepLearningClassification,
  DeepLearningRegression,
  FeatureEmbedding,
  LSTMMultivariateTimeSeries,
  LSTMTimeSeries,
  LogisticRegression,
  MachineLearningLinearRegression,
  MachineLearningLogisticRegression,
  MachineLearningModelInterface,
  MultipleLinearRegression,
  TensorScriptModelInterface,
  TextEmbedding,
  createModelFitCallback,
  getBackend,
  getScikit,
  setBackend,
  setScikit
});
